#! /usr/bin/env python

"""
ProbLog command-line interface.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import sys
import logging
import traceback

from ..program import PrologFile
from ..engine import DefaultEngine
from ..logic import Term, Constant
from ..formula import LogicFormula
from ..errors import process_error
from .. import get_evaluatables, get_evaluatable
from ..util import init_logger, Timer, format_dictionary
from ..program import ExtendedPrologFactory


def main(argv, result_handler=None):
    args = argparser().parse_args(argv)
    inputfile = args.inputfile

    init_logger(args.verbose, name='dtproblog')

    if result_handler is None:
        if args.web:
            result_handler = print_result_json
        else:
            result_handler = print_result

    if args.output is not None:
        outf = open(args.output, 'w')
    else:
        outf = sys.stdout

    try:

        # factory = DTProbLogFactory()
        with Timer('Total', logger='dtproblog'):
            with Timer('Parse input', logger='dtproblog'):
                pl = PrologFile(inputfile)  # factory=factory
                eng = DefaultEngine(label_all=True)
                db = eng.prepare(pl)

                # print (db)

            with Timer('Ground', logger='dtproblog'):
                # decisions = dict((d[0], None) for d in eng.query(db, Term('decision', None)))
                utilities = dict(eng.query(db, Term('utility', None, None)))

                # logging.getLogger('dtproblog').debug('Decisions: %s' % decisions)
                logging.getLogger('dtproblog').debug('Utilities: %s' % utilities)

                # for d in decisions:
                #     db += d.with_probability(Constant(0.5))

                gp = eng.ground_all(db, target=None, queries=utilities.keys())
                decisions = []
                decision_nodes = set()
                for i, n, t in gp:
                    if t == 'atom' and n.probability == Term('?'):
                        decisions.append((i, n.name))
                        decision_nodes.add(i)

                constraints = []
                for c in gp.constraints():
                    if set(c.get_nodes()) & decision_nodes:
                        constraints.append(c)

                # print (gp)

            with Timer('Compile', logger='dtproblog'):
                knowledge = get_evaluatable(args.koption).create_from(gp)

            with Timer('Optimize', logger='dtproblog'):
                if args.search == 'local':
                    result = search_local(knowledge, decisions, utilities, constraints, **vars(args))
                else:
                    result = search_exhaustive(knowledge, decisions, utilities, constraints, **vars(args))

        choices, score, stats = result
        logging.getLogger('dtproblog').info('Number of strategies evaluated: %s' % stats.get('eval'))

        renamed_choices = {}
        for k, v in choices.items():
            if k.functor == 'choice':
                if args.web:
                    k.args[2].loc = db.lineno(k.args[2].location)
                k = k.args[2]
            else:
                if args.web:
                    k.loc = db.lineno(k.location)
            renamed_choices[k] = v

        result_handler((True, (renamed_choices, score, stats)), outf)
    except Exception as err:
        err.trace = traceback.format_exc()
        print (err.trace)
        result_handler((False, err), outf)

    if args.output is not None:
        outf.close()


def evaluate(formula, decisions, utilities):
    result = formula.evaluate(weights=decisions)

    score = 0.0
    for r in result:
        score += result[r] * float(utilities[r])
    return score


def search_exhaustive(formula, decisions, utilities, constraints, verbose=0, **kwargs):
    stats = {'eval': 0}
    best_score = None
    best_choice = None

    decision_ids, decision_names = zip(*decisions)

    # decision_names = decisions.keys()
    for i in range(0, 1 << len(decisions)):
        choices = num2bits(i, len(decisions))

        evidence = dict(zip(decision_names, map(int, choices)))

        constraints_ok = True
        for c in constraints:
            if not c.check(dict(zip(decision_ids, map(int, choices)))):
                constraints_ok = False
                break
        if not constraints_ok:
            continue

        score = evaluate(formula, evidence, utilities)
        stats['eval'] += 1
        if best_score is None or score > best_score:
            best_score = score
            best_choice = dict(evidence)
            logging.getLogger('dtproblog').debug('Improvement: %s -> %s' % (best_choice, best_score))
    return best_choice, best_score, stats


def search_local(formula, decisions, utilities, constraints, verbose=0, **kwargs):
    """Performs local search.

    :param formula:
    :param decisions:
    :param utilities:
    :param verbose:
    :param kwargs:
    :return:
    """
    stats = {'eval': 1}

    # Create the initial strategy:
    #  for each decision, take option that has highest local utility
    #   (takes false if no utility is given for the decision variable)
    for key in decisions:
        if key in utilities and float(utilities[key]) > 0:
            decisions[key] = True
        else:
            decisions[key] = False

    # Compute the score of the initial strategy.
    best_score = evaluate(formula, decisions, utilities)

    # Perform local search by flipping one decision at a time
    last_update = None  # Last decision that was flipped and improved the score
    stop = False
    while not stop:
        # This loop stops when either of these conditions is met:
        #   - at the end of the (first) iteration no decision was flipped successfully
        #   - while iterating we again reach the last decision that was flipped
        #       (this means we tried to flip all decisions, but none were successfull)
        for key in decisions:
            if last_update == key:
                # We went through all decisions without flipping since the last flip.
                stop = True
                break
            # Flip a decision
            decisions[key] = not decisions[key]
            # Compute the score of the new strategy
            flip_score = evaluate(formula, decisions, utilities)
            stats['eval'] += 1
            if flip_score <= best_score:
                # The score is not better: undo the flip
                decisions[key] = not decisions[key]
            else:
                # The score is better: update best score and pointer to last_update
                last_update = key
                best_score = flip_score
                logging.getLogger('dtproblog').debug('Improvement: %s -> %s' % (decisions, best_score))
        if last_update is None:
            # We went through all decisions without flipping.
            stop = True

    return decisions, best_score, stats


def num2bits(n, nbits):
    bits = [False] * nbits
    for i in range(1, nbits + 1):
        bits[nbits - i] = bool(n % 2)
        n >>= 1
    return bits


def print_result(result, output=sys.stdout):
    success, result = result
    if success:
        choices, score, stats = result
        print(format_dictionary(choices, 0), file=output)
        print('SCORE: %s' % score, file=output)
        return 0
    else:
        print(process_error(result), file=output)
        return 1


def print_result_json(d, output):
    """Pretty print result.

    :param d: result from run_problog
    :param output: output file
    :return:
    """
    import json
    result = {}
    success, d = d
    if success:
        choices, score, stats = d
        result['SUCCESS'] = True

        resout = []
        for n, p in choices.items():
            if hasattr(n, 'loc') and n.loc is not None:
                resout.append([str(n), int(p), n.loc[1], n.loc[2]])
            else:
                resout.append([str(n), int(p), None, None])
        result['choices'] = resout
        result['score'] = score
        result['stats'] = stats
    else:
        result['SUCCESS'] = False
        result['err'] = vars(d)
        result['err']['message'] = process_error(result)
    print (json.dumps(result), file=output)
    return 0


# class DTProbLogFactory(ExtendedPrologFactory):
#
#     def build_probabilistic(self, operand1, operand2, location=None, **extra):
#         if str(operand1.functor) in 'd?':
#             return Term('decision', operand2, location=location)
#         else:
#             return ExtendedPrologFactory.build_probabilistic(self, operand1, operand2, location, **extra)


def argparser():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('inputfile')
    parser.add_argument('--knowledge', '-k', dest='koption',
                        choices=get_evaluatables(),
                        default=None, help="Knowledge compilation tool.")
    parser.add_argument('-s', '--search', choices=('local', 'exhaustive'), default='exhaustive')
    parser.add_argument('-v', '--verbose', action='count', help='Set verbosity level')
    parser.add_argument('-o', '--output', type=str, default=None,
                        help='Write output to given file (default: write to stdout)')
    parser.add_argument('--web', action='store_true', help=argparse.SUPPRESS)
    return parser


if __name__ == '__main__':
    main(sys.argv[1:])



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.5
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info
if version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_sdd', [dirname(__file__)])
        except ImportError:
            import _sdd
            return _sdd
        if fp is not None:
            try:
                _mod = imp.load_module('_sdd', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _sdd = swig_import_helper()
    del swig_import_helper
else:
    import _sdd
del version_info
try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr_nondynamic(self, class_type, name, static=1):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    if (not static):
        return object.__getattr__(self, name)
    else:
        raise AttributeError(name)

def _swig_getattr(self, class_type, name):
    return _swig_getattr_nondynamic(self, class_type, name, 0)


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object:
        pass
    _newclass = 0



def sdd_array_element(arr, element):
    return _sdd.sdd_array_element(arr, element)
sdd_array_element = _sdd.sdd_array_element

def sdd_array_int_element(arr, element):
    return _sdd.sdd_array_int_element(arr, element)
sdd_array_int_element = _sdd.sdd_array_int_element

_sdd.PRIsS_swigconstant(_sdd)
PRIsS = _sdd.PRIsS

_sdd.PRInsS_swigconstant(_sdd)
PRInsS = _sdd.PRInsS

_sdd.PRIrcS_swigconstant(_sdd)
PRIrcS = _sdd.PRIrcS

_sdd.PRImcS_swigconstant(_sdd)
PRImcS = _sdd.PRImcS

_sdd.PRIlitS_swigconstant(_sdd)
PRIlitS = _sdd.PRIlitS

_sdd.CONJOIN_swigconstant(_sdd)
CONJOIN = _sdd.CONJOIN

_sdd.DISJOIN_swigconstant(_sdd)
DISJOIN = _sdd.DISJOIN

def sdd_manager_new(vtree):
    return _sdd.sdd_manager_new(vtree)
sdd_manager_new = _sdd.sdd_manager_new

def sdd_manager_create(var_count, auto_gc_and_minimize):
    return _sdd.sdd_manager_create(var_count, auto_gc_and_minimize)
sdd_manager_create = _sdd.sdd_manager_create

def sdd_manager_free(manager):
    return _sdd.sdd_manager_free(manager)
sdd_manager_free = _sdd.sdd_manager_free

def sdd_manager_print(manager):
    return _sdd.sdd_manager_print(manager)
sdd_manager_print = _sdd.sdd_manager_print

def sdd_manager_auto_gc_and_minimize_on(manager):
    return _sdd.sdd_manager_auto_gc_and_minimize_on(manager)
sdd_manager_auto_gc_and_minimize_on = _sdd.sdd_manager_auto_gc_and_minimize_on

def sdd_manager_auto_gc_and_minimize_off(manager):
    return _sdd.sdd_manager_auto_gc_and_minimize_off(manager)
sdd_manager_auto_gc_and_minimize_off = _sdd.sdd_manager_auto_gc_and_minimize_off

def sdd_manager_set_minimize_function(func, manager):
    return _sdd.sdd_manager_set_minimize_function(func, manager)
sdd_manager_set_minimize_function = _sdd.sdd_manager_set_minimize_function

def sdd_manager_unset_minimize_function(manager):
    return _sdd.sdd_manager_unset_minimize_function(manager)
sdd_manager_unset_minimize_function = _sdd.sdd_manager_unset_minimize_function

def sdd_manager_options(manager):
    return _sdd.sdd_manager_options(manager)
sdd_manager_options = _sdd.sdd_manager_options

def sdd_manager_set_options(options, manager):
    return _sdd.sdd_manager_set_options(options, manager)
sdd_manager_set_options = _sdd.sdd_manager_set_options

def sdd_manager_is_var_used(var, manager):
    return _sdd.sdd_manager_is_var_used(var, manager)
sdd_manager_is_var_used = _sdd.sdd_manager_is_var_used

def sdd_manager_vtree_of_var(var, manager):
    return _sdd.sdd_manager_vtree_of_var(var, manager)
sdd_manager_vtree_of_var = _sdd.sdd_manager_vtree_of_var

def sdd_manager_lca_of_literals(count, literals, manager):
    return _sdd.sdd_manager_lca_of_literals(count, literals, manager)
sdd_manager_lca_of_literals = _sdd.sdd_manager_lca_of_literals

def sdd_manager_var_count(manager):
    return _sdd.sdd_manager_var_count(manager)
sdd_manager_var_count = _sdd.sdd_manager_var_count

def sdd_manager_var_order(var_order, manager):
    return _sdd.sdd_manager_var_order(var_order, manager)
sdd_manager_var_order = _sdd.sdd_manager_var_order

def sdd_manager_add_var_before_first(manager):
    return _sdd.sdd_manager_add_var_before_first(manager)
sdd_manager_add_var_before_first = _sdd.sdd_manager_add_var_before_first

def sdd_manager_add_var_after_last(manager):
    return _sdd.sdd_manager_add_var_after_last(manager)
sdd_manager_add_var_after_last = _sdd.sdd_manager_add_var_after_last

def sdd_manager_add_var_before(target_var, manager):
    return _sdd.sdd_manager_add_var_before(target_var, manager)
sdd_manager_add_var_before = _sdd.sdd_manager_add_var_before

def sdd_manager_add_var_after(target_var, manager):
    return _sdd.sdd_manager_add_var_after(target_var, manager)
sdd_manager_add_var_after = _sdd.sdd_manager_add_var_after

def sdd_manager_true(manager):
    return _sdd.sdd_manager_true(manager)
sdd_manager_true = _sdd.sdd_manager_true

def sdd_manager_false(manager):
    return _sdd.sdd_manager_false(manager)
sdd_manager_false = _sdd.sdd_manager_false

def sdd_manager_literal(literal, manager):
    return _sdd.sdd_manager_literal(literal, manager)
sdd_manager_literal = _sdd.sdd_manager_literal

def sdd_apply(node1, node2, op, manager):
    return _sdd.sdd_apply(node1, node2, op, manager)
sdd_apply = _sdd.sdd_apply

def sdd_apply_in_vtree(node1, node2, op, vtree, manager):
    return _sdd.sdd_apply_in_vtree(node1, node2, op, vtree, manager)
sdd_apply_in_vtree = _sdd.sdd_apply_in_vtree

def sdd_conjoin(node1, node2, manager):
    return _sdd.sdd_conjoin(node1, node2, manager)
sdd_conjoin = _sdd.sdd_conjoin

def sdd_disjoin(node1, node2, manager):
    return _sdd.sdd_disjoin(node1, node2, manager)
sdd_disjoin = _sdd.sdd_disjoin

def sdd_negate(node, manager):
    return _sdd.sdd_negate(node, manager)
sdd_negate = _sdd.sdd_negate

def sdd_condition(lit, node, manager):
    return _sdd.sdd_condition(lit, node, manager)
sdd_condition = _sdd.sdd_condition

def sdd_exists(var, node, manager):
    return _sdd.sdd_exists(var, node, manager)
sdd_exists = _sdd.sdd_exists

def sdd_forall(var, node, manager):
    return _sdd.sdd_forall(var, node, manager)
sdd_forall = _sdd.sdd_forall

def sdd_minimize_cardinality(node, manager):
    return _sdd.sdd_minimize_cardinality(node, manager)
sdd_minimize_cardinality = _sdd.sdd_minimize_cardinality

def sdd_minimum_cardinality(node):
    return _sdd.sdd_minimum_cardinality(node)
sdd_minimum_cardinality = _sdd.sdd_minimum_cardinality

def sdd_model_count(node, manager):
    return _sdd.sdd_model_count(node, manager)
sdd_model_count = _sdd.sdd_model_count

def sdd_node_is_true(node):
    return _sdd.sdd_node_is_true(node)
sdd_node_is_true = _sdd.sdd_node_is_true

def sdd_node_is_false(node):
    return _sdd.sdd_node_is_false(node)
sdd_node_is_false = _sdd.sdd_node_is_false

def sdd_node_is_literal(node):
    return _sdd.sdd_node_is_literal(node)
sdd_node_is_literal = _sdd.sdd_node_is_literal

def sdd_node_is_decision(node):
    return _sdd.sdd_node_is_decision(node)
sdd_node_is_decision = _sdd.sdd_node_is_decision

def sdd_node_size(node):
    return _sdd.sdd_node_size(node)
sdd_node_size = _sdd.sdd_node_size

def sdd_node_literal(node):
    return _sdd.sdd_node_literal(node)
sdd_node_literal = _sdd.sdd_node_literal

def sdd_node_elements(node):
    return _sdd.sdd_node_elements(node)
sdd_node_elements = _sdd.sdd_node_elements

def sdd_node_set_bit(bit, node):
    return _sdd.sdd_node_set_bit(bit, node)
sdd_node_set_bit = _sdd.sdd_node_set_bit

def sdd_node_bit(node):
    return _sdd.sdd_node_bit(node)
sdd_node_bit = _sdd.sdd_node_bit

def sdd_id(node):
    return _sdd.sdd_id(node)
sdd_id = _sdd.sdd_id

def sdd_garbage_collected(node, id):
    return _sdd.sdd_garbage_collected(node, id)
sdd_garbage_collected = _sdd.sdd_garbage_collected

def sdd_vtree_of(node):
    return _sdd.sdd_vtree_of(node)
sdd_vtree_of = _sdd.sdd_vtree_of

def sdd_copy(node, dest_manager):
    return _sdd.sdd_copy(node, dest_manager)
sdd_copy = _sdd.sdd_copy

def sdd_rename_variables(node, variable_map, manager):
    return _sdd.sdd_rename_variables(node, variable_map, manager)
sdd_rename_variables = _sdd.sdd_rename_variables

def sdd_variables(node, manager):
    return _sdd.sdd_variables(node, manager)
sdd_variables = _sdd.sdd_variables

def sdd_read(filename, manager):
    return _sdd.sdd_read(filename, manager)
sdd_read = _sdd.sdd_read

def sdd_save(fname, node):
    return _sdd.sdd_save(fname, node)
sdd_save = _sdd.sdd_save

def sdd_save_as_dot(fname, node):
    return _sdd.sdd_save_as_dot(fname, node)
sdd_save_as_dot = _sdd.sdd_save_as_dot

def sdd_shared_save_as_dot(fname, manager):
    return _sdd.sdd_shared_save_as_dot(fname, manager)
sdd_shared_save_as_dot = _sdd.sdd_shared_save_as_dot

def sdd_count(node):
    return _sdd.sdd_count(node)
sdd_count = _sdd.sdd_count

def sdd_size(node):
    return _sdd.sdd_size(node)
sdd_size = _sdd.sdd_size

def sdd_shared_size(nodes, count):
    return _sdd.sdd_shared_size(nodes, count)
sdd_shared_size = _sdd.sdd_shared_size

def sdd_manager_size(manager):
    return _sdd.sdd_manager_size(manager)
sdd_manager_size = _sdd.sdd_manager_size

def sdd_manager_live_size(manager):
    return _sdd.sdd_manager_live_size(manager)
sdd_manager_live_size = _sdd.sdd_manager_live_size

def sdd_manager_dead_size(manager):
    return _sdd.sdd_manager_dead_size(manager)
sdd_manager_dead_size = _sdd.sdd_manager_dead_size

def sdd_manager_count(manager):
    return _sdd.sdd_manager_count(manager)
sdd_manager_count = _sdd.sdd_manager_count

def sdd_manager_live_count(manager):
    return _sdd.sdd_manager_live_count(manager)
sdd_manager_live_count = _sdd.sdd_manager_live_count

def sdd_manager_dead_count(manager):
    return _sdd.sdd_manager_dead_count(manager)
sdd_manager_dead_count = _sdd.sdd_manager_dead_count

def sdd_vtree_size(vtree):
    return _sdd.sdd_vtree_size(vtree)
sdd_vtree_size = _sdd.sdd_vtree_size

def sdd_vtree_live_size(vtree):
    return _sdd.sdd_vtree_live_size(vtree)
sdd_vtree_live_size = _sdd.sdd_vtree_live_size

def sdd_vtree_dead_size(vtree):
    return _sdd.sdd_vtree_dead_size(vtree)
sdd_vtree_dead_size = _sdd.sdd_vtree_dead_size

def sdd_vtree_count(vtree):
    return _sdd.sdd_vtree_count(vtree)
sdd_vtree_count = _sdd.sdd_vtree_count

def sdd_vtree_live_count(vtree):
    return _sdd.sdd_vtree_live_count(vtree)
sdd_vtree_live_count = _sdd.sdd_vtree_live_count

def sdd_vtree_dead_count(vtree):
    return _sdd.sdd_vtree_dead_count(vtree)
sdd_vtree_dead_count = _sdd.sdd_vtree_dead_count

def sdd_vtree_size_at(vtree):
    return _sdd.sdd_vtree_size_at(vtree)
sdd_vtree_size_at = _sdd.sdd_vtree_size_at

def sdd_vtree_live_size_at(vtree):
    return _sdd.sdd_vtree_live_size_at(vtree)
sdd_vtree_live_size_at = _sdd.sdd_vtree_live_size_at

def sdd_vtree_dead_size_at(vtree):
    return _sdd.sdd_vtree_dead_size_at(vtree)
sdd_vtree_dead_size_at = _sdd.sdd_vtree_dead_size_at

def sdd_vtree_count_at(vtree):
    return _sdd.sdd_vtree_count_at(vtree)
sdd_vtree_count_at = _sdd.sdd_vtree_count_at

def sdd_vtree_live_count_at(vtree):
    return _sdd.sdd_vtree_live_count_at(vtree)
sdd_vtree_live_count_at = _sdd.sdd_vtree_live_count_at

def sdd_vtree_dead_count_at(vtree):
    return _sdd.sdd_vtree_dead_count_at(vtree)
sdd_vtree_dead_count_at = _sdd.sdd_vtree_dead_count_at

def sdd_vtree_new(var_count, type):
    return _sdd.sdd_vtree_new(var_count, type)
sdd_vtree_new = _sdd.sdd_vtree_new

def sdd_vtree_new_with_var_order(var_count, var_order, type):
    return _sdd.sdd_vtree_new_with_var_order(var_count, var_order, type)
sdd_vtree_new_with_var_order = _sdd.sdd_vtree_new_with_var_order

def sdd_vtree_free(vtree):
    return _sdd.sdd_vtree_free(vtree)
sdd_vtree_free = _sdd.sdd_vtree_free

def sdd_vtree_save(fname, vtree):
    return _sdd.sdd_vtree_save(fname, vtree)
sdd_vtree_save = _sdd.sdd_vtree_save

def sdd_vtree_read(filename):
    return _sdd.sdd_vtree_read(filename)
sdd_vtree_read = _sdd.sdd_vtree_read

def sdd_vtree_save_as_dot(fname, vtree):
    return _sdd.sdd_vtree_save_as_dot(fname, vtree)
sdd_vtree_save_as_dot = _sdd.sdd_vtree_save_as_dot

def sdd_manager_vtree(manager):
    return _sdd.sdd_manager_vtree(manager)
sdd_manager_vtree = _sdd.sdd_manager_vtree

def sdd_manager_vtree_copy(manager):
    return _sdd.sdd_manager_vtree_copy(manager)
sdd_manager_vtree_copy = _sdd.sdd_manager_vtree_copy

def sdd_vtree_left(vtree):
    return _sdd.sdd_vtree_left(vtree)
sdd_vtree_left = _sdd.sdd_vtree_left

def sdd_vtree_right(vtree):
    return _sdd.sdd_vtree_right(vtree)
sdd_vtree_right = _sdd.sdd_vtree_right

def sdd_vtree_parent(vtree):
    return _sdd.sdd_vtree_parent(vtree)
sdd_vtree_parent = _sdd.sdd_vtree_parent

def sdd_vtree_is_leaf(vtree):
    return _sdd.sdd_vtree_is_leaf(vtree)
sdd_vtree_is_leaf = _sdd.sdd_vtree_is_leaf

def sdd_vtree_is_sub(vtree1, vtree2):
    return _sdd.sdd_vtree_is_sub(vtree1, vtree2)
sdd_vtree_is_sub = _sdd.sdd_vtree_is_sub

def sdd_vtree_lca(vtree1, vtree2, root):
    return _sdd.sdd_vtree_lca(vtree1, vtree2, root)
sdd_vtree_lca = _sdd.sdd_vtree_lca

def sdd_vtree_var_count(vtree):
    return _sdd.sdd_vtree_var_count(vtree)
sdd_vtree_var_count = _sdd.sdd_vtree_var_count

def sdd_vtree_var(vtree):
    return _sdd.sdd_vtree_var(vtree)
sdd_vtree_var = _sdd.sdd_vtree_var

def sdd_vtree_position(vtree):
    return _sdd.sdd_vtree_position(vtree)
sdd_vtree_position = _sdd.sdd_vtree_position

def sdd_vtree_location(vtree, manager):
    return _sdd.sdd_vtree_location(vtree, manager)
sdd_vtree_location = _sdd.sdd_vtree_location

def sdd_vtree_rotate_left(vtree, manager, time_limit, size_limit):
    return _sdd.sdd_vtree_rotate_left(vtree, manager, time_limit, size_limit)
sdd_vtree_rotate_left = _sdd.sdd_vtree_rotate_left

def sdd_vtree_rotate_right(vtree, manager, time_limit, size_limit, cartesian_product_limit):
    return _sdd.sdd_vtree_rotate_right(vtree, manager, time_limit, size_limit, cartesian_product_limit)
sdd_vtree_rotate_right = _sdd.sdd_vtree_rotate_right

def sdd_vtree_swap(vtree, manager, time_limit, size_limit, cartesian_product_limit):
    return _sdd.sdd_vtree_swap(vtree, manager, time_limit, size_limit, cartesian_product_limit)
sdd_vtree_swap = _sdd.sdd_vtree_swap

def sdd_manager_set_size_limit_context(vtree, manager):
    return _sdd.sdd_manager_set_size_limit_context(vtree, manager)
sdd_manager_set_size_limit_context = _sdd.sdd_manager_set_size_limit_context

def sdd_manager_update_size_limit_context(manager):
    return _sdd.sdd_manager_update_size_limit_context(manager)
sdd_manager_update_size_limit_context = _sdd.sdd_manager_update_size_limit_context

def sdd_vtree_bit(vtree):
    return _sdd.sdd_vtree_bit(vtree)
sdd_vtree_bit = _sdd.sdd_vtree_bit

def sdd_vtree_set_bit(bit, vtree):
    return _sdd.sdd_vtree_set_bit(bit, vtree)
sdd_vtree_set_bit = _sdd.sdd_vtree_set_bit

def sdd_vtree_data(vtree):
    return _sdd.sdd_vtree_data(vtree)
sdd_vtree_data = _sdd.sdd_vtree_data

def sdd_vtree_set_data(data, vtree):
    return _sdd.sdd_vtree_set_data(data, vtree)
sdd_vtree_set_data = _sdd.sdd_vtree_set_data

def sdd_vtree_search_state(vtree):
    return _sdd.sdd_vtree_search_state(vtree)
sdd_vtree_search_state = _sdd.sdd_vtree_search_state

def sdd_vtree_set_search_state(search_state, vtree):
    return _sdd.sdd_vtree_set_search_state(search_state, vtree)
sdd_vtree_set_search_state = _sdd.sdd_vtree_set_search_state

def sdd_ref_count(node):
    return _sdd.sdd_ref_count(node)
sdd_ref_count = _sdd.sdd_ref_count

def sdd_ref(node, manager):
    return _sdd.sdd_ref(node, manager)
sdd_ref = _sdd.sdd_ref

def sdd_deref(node, manager):
    return _sdd.sdd_deref(node, manager)
sdd_deref = _sdd.sdd_deref

def sdd_manager_garbage_collect(manager):
    return _sdd.sdd_manager_garbage_collect(manager)
sdd_manager_garbage_collect = _sdd.sdd_manager_garbage_collect

def sdd_vtree_garbage_collect(vtree, manager):
    return _sdd.sdd_vtree_garbage_collect(vtree, manager)
sdd_vtree_garbage_collect = _sdd.sdd_vtree_garbage_collect

def sdd_manager_garbage_collect_if(dead_node_threshold, manager):
    return _sdd.sdd_manager_garbage_collect_if(dead_node_threshold, manager)
sdd_manager_garbage_collect_if = _sdd.sdd_manager_garbage_collect_if

def sdd_vtree_garbage_collect_if(dead_node_threshold, vtree, manager):
    return _sdd.sdd_vtree_garbage_collect_if(dead_node_threshold, vtree, manager)
sdd_vtree_garbage_collect_if = _sdd.sdd_vtree_garbage_collect_if

def sdd_manager_minimize(manager):
    return _sdd.sdd_manager_minimize(manager)
sdd_manager_minimize = _sdd.sdd_manager_minimize

def sdd_vtree_minimize(vtree, manager):
    return _sdd.sdd_vtree_minimize(vtree, manager)
sdd_vtree_minimize = _sdd.sdd_vtree_minimize

def sdd_manager_set_lr_time_limit(time_limit, manager):
    return _sdd.sdd_manager_set_lr_time_limit(time_limit, manager)
sdd_manager_set_lr_time_limit = _sdd.sdd_manager_set_lr_time_limit

def sdd_manager_set_rr_time_limit(time_limit, manager):
    return _sdd.sdd_manager_set_rr_time_limit(time_limit, manager)
sdd_manager_set_rr_time_limit = _sdd.sdd_manager_set_rr_time_limit

def sdd_manager_set_sw_time_limit(time_limit, manager):
    return _sdd.sdd_manager_set_sw_time_limit(time_limit, manager)
sdd_manager_set_sw_time_limit = _sdd.sdd_manager_set_sw_time_limit

def sdd_manager_set_lr_size_limit(size_limit, manager):
    return _sdd.sdd_manager_set_lr_size_limit(size_limit, manager)
sdd_manager_set_lr_size_limit = _sdd.sdd_manager_set_lr_size_limit

def sdd_manager_set_rr_size_limit(size_limit, manager):
    return _sdd.sdd_manager_set_rr_size_limit(size_limit, manager)
sdd_manager_set_rr_size_limit = _sdd.sdd_manager_set_rr_size_limit

def sdd_manager_set_sw_size_limit(size_limit, manager):
    return _sdd.sdd_manager_set_sw_size_limit(size_limit, manager)
sdd_manager_set_sw_size_limit = _sdd.sdd_manager_set_sw_size_limit

def sdd_manager_set_rr_cartesian_product_limit(cartesian_product_limit, manager):
    return _sdd.sdd_manager_set_rr_cartesian_product_limit(cartesian_product_limit, manager)
sdd_manager_set_rr_cartesian_product_limit = _sdd.sdd_manager_set_rr_cartesian_product_limit

def sdd_manager_set_sw_cartesian_product_limit(cartesian_product_limit, manager):
    return _sdd.sdd_manager_set_sw_cartesian_product_limit(cartesian_product_limit, manager)
sdd_manager_set_sw_cartesian_product_limit = _sdd.sdd_manager_set_sw_cartesian_product_limit

def sdd_manager_set_convergence_threshold(threshold, manager):
    return _sdd.sdd_manager_set_convergence_threshold(threshold, manager)
sdd_manager_set_convergence_threshold = _sdd.sdd_manager_set_convergence_threshold

def wmc_manager_new(node, log_mode, manager):
    return _sdd.wmc_manager_new(node, log_mode, manager)
wmc_manager_new = _sdd.wmc_manager_new

def wmc_manager_free(wmc_manager):
    return _sdd.wmc_manager_free(wmc_manager)
wmc_manager_free = _sdd.wmc_manager_free

def wmc_set_literal_weight(literal, weight, wmc_manager):
    return _sdd.wmc_set_literal_weight(literal, weight, wmc_manager)
wmc_set_literal_weight = _sdd.wmc_set_literal_weight

def wmc_propagate(wmc_manager):
    return _sdd.wmc_propagate(wmc_manager)
wmc_propagate = _sdd.wmc_propagate

def wmc_zero_weight(wmc_manager):
    return _sdd.wmc_zero_weight(wmc_manager)
wmc_zero_weight = _sdd.wmc_zero_weight

def wmc_one_weight(wmc_manager):
    return _sdd.wmc_one_weight(wmc_manager)
wmc_one_weight = _sdd.wmc_one_weight

def wmc_literal_weight(literal, wmc_manager):
    return _sdd.wmc_literal_weight(literal, wmc_manager)
wmc_literal_weight = _sdd.wmc_literal_weight

def wmc_literal_derivative(literal, wmc_manager):
    return _sdd.wmc_literal_derivative(literal, wmc_manager)
wmc_literal_derivative = _sdd.wmc_literal_derivative

def wmc_literal_pr(literal, wmc_manager):
    return _sdd.wmc_literal_pr(literal, wmc_manager)
wmc_literal_pr = _sdd.wmc_literal_pr
# This file is compatible with both classic and new-style classes.





#!/usr/bin/env python
# encoding: utf-8
"""Simple HTTP server for ProbLog.

This module only defines the server interface.
To change settings related to ProbLog, you should consult the run_problog.py script.

The operation of the server is determined by four options:

    port (default: 8000)        Server port to use
    timeout (default: 60)       Maximum *processing* time of the ProbLog subprocess
    memout (default: 1Gb)       Maximum memory usage of the ProbLog subprocess
    servefiles (default: No)    Whether to serve a file for undefined paths. (This is potentially unsafe.)

The server defines the following paths:

    http://hostname:port/problog?model=... [GET,POST]
    http://hostname:port/inference?model=...&callback=... [JSONP]
    http://hostname:port/learn?model=...&examples=...&callback=... [JSONP]

If ``servefiles`` is enabled other paths will be treated as file access requests.

__author__ = Anton Dries
__author__ = Wannes Meert

Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import print_function

# Load general Python modules
import sys
import os
import tempfile
import traceback
import subprocess
import resource
import logging
import logging.config

DEFAULT_PORT = 5100
DEFAULT_TIMEOUT = 60
DEFAULT_MEMOUT = 2.0  # gigabyte

SERVE_FILES = False
CACHE_MODELS = True
CACHE_DIR = "cache"

RUN_LOCAL = False

# PYTHON_EXEC = 'python'    # Python 2
PYTHON_EXEC = sys.executable  # Match with server

api_root = '/'

FILES_WHITELIST = [
    api_root + 'js/problog_editor_advanced.js',
    api_root + 'js/lib/ace.js',
    api_root + 'js/lib/bootstrap-theme.min.css',
    api_root + 'js/lib/bootstrap.min.css',
    api_root + 'js/lib/bootstrap.min.js',
    api_root + 'js/lib/jquery-2.1.0.min.js',
    api_root + 'js/lib/jquery-ui.min.js',
    api_root + 'js/lib/mode-prolog.js',
    api_root + 'tutorial.html',
    api_root + 'sidebar.css'
]

here = os.path.dirname(__file__)

try :
    logging.config.fileConfig(os.path.join(here,'logging.conf'))
except IOError:
    logger = logging.getLogger('server')
    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('[%(levelname)s] %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.setLevel(logging.DEBUG)
logger = logging.getLogger('server')

# Load Python standard web-related modules (based on Python version)
import json
import mimetypes
if sys.version_info.major == 2:
    import BaseHTTPServer
    import urlparse

    def to_bytes(string):
        return bytes(string)
    import urllib
    url_decode = urllib.unquote_plus
    import hashlib

    def compute_hash(model):
        return hashlib.md5(model).hexdigest()  # Python 2

else:
    import http.server as BaseHTTPServer
    import urllib.parse as urlparse

    def to_bytes(string):
        return bytes(string, 'UTF-8')
    url_decode = urlparse.unquote_plus
    import hashlib

    def compute_hash(model):
        return hashlib.md5(to_bytes(model)).hexdigest() # Python 3

# Contains special URL paths. Initialize with @handle_url
PATHS = {}


class handle_url(object) :
    """Decorator for adding a handler for a URL.

    Example: to add a handler for GET /hello?name=World

    @handle_url('/hello')
    def hello(name) :
        return 200, 'text/plain', 'Hello %s!' % name
    """

    def __init__(self, path) :
        self.path = path

    def __call__(self, f) :
        PATHS[self.path] = f


def root_path(*relative_path):
    """Translate URL path to file system path."""
    return os.path.abspath(os.path.join(os.path.dirname(__file__), *relative_path))

PROB_EXEC = root_path('..', 'tasks', '__init__.py')


def call_process(cmd, timeout, memout):
    """Call a subprocess with time and memory restrictions.

        Note:
            - the time restriction only works on Linux and Mac OS X
            - the memory restriction only works on Linux

        :param:timeout: CPU time in seconds
        :param:memout: Maximum memory in bytes
        :return: Output of the subprocess.

    """

    def setlimits():
        resource.setrlimit(resource.RLIMIT_CPU, (timeout, timeout))
        resource.setrlimit(resource.RLIMIT_AS, (memout, memout))

    return subprocess.check_output(cmd, preexec_fn=setlimits)


def wrap_callback(callback, jsonstr):
    return '{}({});'.format(callback, jsonstr)


def run_problog_task(task, model, callback=None, data=None, options=None):
    if options is None:
        options = []

    try:
        # Write the model to a temporary or cached file.
        infile = store_hash(model, 'pl')

        # Construct the basic command
        cmd = [PYTHON_EXEC, PROB_EXEC, task, infile]

        # Write the data to a temporary or cached file (if given)
        datafile = None
        if data is not None:
            datafile = store_hash(data, 'data')
            outfile = os.path.splitext(infile)[0] + '_' + \
                os.path.splitext(os.path.basename(datafile))[0] + '.out'
            cmd += [datafile]
        else:
            outfile = os.path.splitext(infile)[0] + '.out'
        logger.info('Output file: {}'.format(outfile))

        # Add default and given options to the command.
        cmd += ['-o', outfile, '--web'] + options

    except UnicodeDecodeError as err:
        logger.error('Unicode error catched: {}'.format(err))
        result = {'SUCCESS': False, 'err': 'Cannot decode character in program: {}'.format(err)}
        return 200, 'application/json', wrap_callback(callback, json.dumps(result))

    if CACHE_MODELS:
        url = 'task=%s' % task
        url += '&hash=%s' % os.path.splitext(os.path.basename(infile))[0]
        if data is not None:
            url += '&ehash=%s' % os.path.splitext(os.path.basename(datafile))[0]
    else:
        url = None

    try:
        # Execute ProbLog
        call_process(cmd, DEFAULT_TIMEOUT, DEFAULT_MEMOUT * (1 << 30))

        # Read output produced by ProbLog
        with open(outfile) as f:
            result = f.read()

        if url is not None:
            result = json.loads(result)
            result['url'] = url
            result = json.dumps(result)

        # Wrap the output in a JSON wrapper.
        datavalue = wrap_callback(callback, result)
        return 200, 'application/json', datavalue
    except subprocess.CalledProcessError as err:
        logger.error('ProbLog didn\'t finish correctly: %s' % err)
        result = {'SUCCESS': False, 'url': url,
                  'err': 'ProbLog learning exceeded time or memory limit'}
        return 200, 'application/json', wrap_callback(callback, json.dumps(result))


@handle_url(api_root + 'inference')
def run_problog_jsonp(model, callback):
    return run_problog_task('prob', model[0], callback[0])


@handle_url(api_root + 'prob')
def run_problog_jsonp(model, callback):
    return run_problog_task('prob', model[0], callback[0])


@handle_url(api_root + 'dt')
def run_problog_jsonp(model, callback, solve=None):
    if solve is not None and solve[0] == 'local':
        options = ['--search', 'local']
    else:
        options = []
    return run_problog_task('dt', model[0], callback[0], options=options)


@handle_url(api_root + 'mpe')
def run_mpe_jsonp(model, callback):
    return run_problog_task('mpe', model[0], callback[0], options=['--full'])


@handle_url(api_root + 'sample')
def run_sample_jsonp(model, callback):
    return run_problog_task('sample', model[0], callback[0])


@handle_url(api_root + 'explain')
def run_sample_jsonp(model, callback):
    return run_problog_task('explain', model[0], callback[0])


@handle_url(api_root + 'lfi')
def run_lfi_jsonp(model, examples, callback):
    return run_problog_task('lfi', model[0], callback[0], data=examples[0])


@handle_url(api_root + 'learning')
def run_lfi_jsonp(model, examples, callback):
    return run_problog_task('lfi', model[0], callback[0], data=examples[0])


@handle_url(api_root + 'ground')
def run_lfi_jsonp(model, callback):
    return run_problog_task('ground', model[0], callback[0], options=['--format', 'svg'])


def store_hash(data, datatype):
    # Can raise UnicodeDecodeError
    if CACHE_MODELS:
        datahash = compute_hash(data)
        if not os.path.exists(CACHE_DIR):
            os.mkdir(CACHE_DIR)
        datafile = os.path.join(CACHE_DIR, datahash + '.' + datatype)
        logger.info('Saved file: {}'.format(datafile))
    else:
        _, datafile = tempfile.mkstemp('.' + datatype)

    with open(datafile, 'w') as f:
        f.write(data)

    return datafile


def get_hash(hashcode, datatype):
    datafile = os.path.join(CACHE_DIR, hashcode + '.' + datatype)
    if not CACHE_MODELS or not os.path.exists(datafile):
        return None
    else:
        with open(datafile, 'r') as f:
            data = f.read()
        return data


@handle_url(api_root + 'model')
def get_model_from_hash_jsonp(hash, callback):
    data = get_hash(hash[0], 'pl')
    if data is None:
        result = {'SUCCESS': False, 'err': 'Model hash not available: {}'.format(hash[0])}
    else:
        result = {'SUCCESS': True, 'model': data}
    return 200, 'application/json', wrap_callback(callback[0], json.dumps(result))


@handle_url(api_root + 'examples')
def get_data_from_hash_jsonp(hash, callback):
    data = get_hash(hash[0], 'data')
    if data is None:
        result = {'SUCCESS': False, 'err': 'Model hash not available: {}'.format(hash[0])}
    else:
        result = {'SUCCESS': True, 'examples': data}
    return 200, 'application/json', wrap_callback(callback[0], json.dumps(result))


@handle_url(api_root + 'models')
def get_example_models(callback):
    return


@handle_url(api_root)
def get_editor():
    with open(root_path('editor_adv.html')) as f:
        data = f.read()
    data = data.replace('problog.init()', 'problog.init(\'\');')
    data = make_local(data)
    return 200, 'text/html', data


def make_local(html):
    import re

    if RUN_LOCAL:
        html = re.sub('<link rel="stylesheet" href="(http([^/"]*/)+)', '<link rel="stylesheet" href="js/lib/', html)
        html = re.sub('<script src="(http([^/"]*/)+)', '<script src="js/lib/', html)
        return html
    else:
        return html


# @handle_url(api_root + 'js/problog_editor_advanced.js')
# def get_javascript():
#     with open(root_path('js/problog_editor_advanced.js')) as f:
#         data = f.read()
#     return 200, 'application/javascript', data
#
#
# @handle_url(api_root + 'tutorial.html')
# def get_javascript():
#     with open(root_path('js/problog_editor_advanced.js')) as f:
#         data = f.read()
#     return 200, 'application/javascript', data

class ProbLogHTTP(BaseHTTPServer.BaseHTTPRequestHandler):

    def do_POST(self) :
        numberOfBytes = int(self.headers["Content-Length"])
        data = self.rfile.read(numberOfBytes)
        query = urlparse.parse_qs(data)

        self.do_GET(query=query)

    def do_GET(self, query=None) :
        """Handle a GET request.

        Looks up the URL path in PATHS and executes the corresponding function.
        If the path does not occur in PATHS, treats the path as a filename and serves the file.
        """

        try:
            url = urlparse.urlparse( self.path )
            path = url.path
            if query is None:
                query = urlparse.parse_qs(url.query)
            if '_' in query:
                # Used by jquery to avoid caching
                del query['_']

            action = PATHS.get(path)
            if action is None:
                if SERVE_FILES or str(path) in FILES_WHITELIST:
                    self.serveFile(path)
                else:
                    self.send_response(404)
                    self.end_headers()
                    self.wfile.write(to_bytes('File not found!'))
            else :
                code, datatype, data = action(**query)
                self.send_response(code)
                self.send_header("Content-type", datatype)
                self.end_headers()
                if data:
                    self.wfile.write(to_bytes(data))
        except Exception as e:
            import traceback
            logger.error('Uncaught exception: {}\n{}'.format(e, traceback.format_exc()))

    def serveFile(self, filename) :
        """Serve a file."""
        if filename == '/':
            filename = '/index.html'
        filename = filename.lstrip('/')

        filetype, encoding = mimetypes.guess_type(filename)
        filename = root_path(filename)
        try :
            with open(filename) as f:
                data = f.read()
            data = make_local(data)
            self.send_response(200)
            self.send_header("Content-type", filetype)
            if encoding :
                self.send_header("Content-Encoding", encoding)
            self.end_headers()
            self.wfile.write(to_bytes(data))
        except :
            self.send_response(404)
            self.end_headers()
            self.wfile.write(to_bytes('File not found!'))

    def log_message(self,format, *args) :
        try :
            # Remove arguments from GET request, only keep path
            if args[0].startswith('GET') :
                p = args[0].find('?')
                if p >= 0 :
                    args0 = args[0][0:p]
                else :
                    args0 = args[0]
                args = (args0,) + args[1:]
        except Exception :
            pass

        args = (self.client_address[0],) + args
        format = '[%s] ' + format
        logger.info(format % args)


def main(argv, **extra):
    global DEFAULT_MEMOUT
    global DEFAULT_TIMEOUT
    global SERVE_FILES
    global CACHE_MODELS
    global RUN_LOCAL

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--port', '-p', type=int, default=DEFAULT_PORT, help="Server listening port")
    parser.add_argument('--timeout', '-t', type=int, default=DEFAULT_TIMEOUT, help="Time limit in seconds")
    parser.add_argument('--memout', '-m', type=float, default=DEFAULT_MEMOUT, help="Memory limit in Gb")
    parser.add_argument('--servefiles', '-F', action='store_true', help="Attempt to serve a file for undefined paths (unsafe?).")
    parser.add_argument('--nocaching', action='store_true', help="Disable caching of submitted models")
    parser.add_argument('--browser', '-B', action='store_true', help="Open editor in web browser.")
    parser.add_argument('--local', '-l', action='store_true', help="Use local javascript libraries.")
    args = parser.parse_args(argv)

    RUN_LOCAL = args.local
    if args.local:
        args.servefiles = True

    DEFAULT_TIMEOUT = args.timeout
    DEFAULT_MEMOUT = args.memout
    SERVE_FILES = args.servefiles
    CACHE_MODELS = not args.nocaching
    logger.info('Starting server on port %d (timeout=%d, memout=%dGb)' % (args.port, DEFAULT_TIMEOUT, DEFAULT_MEMOUT))

    server_address = ('', args.port)
    httpd = BaseHTTPServer.HTTPServer(server_address, ProbLogHTTP)
    if args.browser:
        import webbrowser
        webbrowser.open('http://localhost:%s/' % args.port, new=2, autoraise=True)
    httpd.serve_forever()


if __name__ == '__main__':
    main(sys.argv[1:])



"""
problog.engine_unify - Unification
----------------------------------

Implementation of unification for the grounding engine.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function
from .errors import GroundingError
from .logic import is_variable


class UnifyError(Exception):
    """Unification error (used and handled internally)."""
    pass


class _SubstitutionWrapper(object):
    def __init__(self, subst):
        self.subst = subst

    def __getitem__(self, key):
        if key in self.subst:
            return self.subst[key]
        else:
            return key


def substitute_all(terms, subst):
    """

    :param terms:
    :param subst:
    :return:
    """
    subst = _SubstitutionWrapper(subst)
    result = []
    for term in terms:
        if is_variable(term):
            result.append(subst[term])
        else:
            result.append(term.apply(subst))
    return result


def instantiate(term, context):
    """Replace variables in Term by values based on context lookup table.

    :param term:
    :param context:
    :return:
    """
    if is_variable(term):
        return context[term]
    else:
        return term.apply(context)


class OccursCheck(GroundingError):
    def __init__(self):
        GroundingError.__init__(self, 'Infinite unification')


def unify_value(value1, value2, source_values):
    """
    Unify two values that exist in the same context.
    :param value1:
    :param value2:
    :param source_values:
    :return:
    """
    # Variables are negative numbers or None
    # Naive implementation (no occurs check)

    if is_variable(value1) and is_variable(value2):
        if value1 == value2:
            return value1
        elif value1 is None:
            return value2
        if value2 is None:  # second one is anonymous
            return value1
        else:
            # Two named variables: unify their source_values
            value = unify_value(source_values.get(value1), source_values.get(value2), source_values)
            if value is None:
                value = max(value1, value2)
            source_values[value1] = value
            source_values[value2] = value
            # Return the one of lowest rank: negative numbers, so max
            return value
    elif is_variable(value1):
        if value1 is None:
            return value2
        else:
            if value1 in value2.variables():
                raise OccursCheck()
            value = unify_value(source_values.get(value1), value2, source_values)
            source_values[value1] = value
            return value
    elif is_variable(value2):
        if value2 is None:
            return value1
        else:
            if value2 in value1.variables():
                raise OccursCheck()
            value = unify_value(source_values.get(value2), value1, source_values)
            source_values[value2] = value
            return value
    elif value1.signature == value2.signature:  # Assume Term
        return value1.with_args(*[unify_value(a1, a2, source_values)
                                  for a1, a2 in zip(value1.args, value2.args)])
    else:
        raise UnifyError()


class _ContextWrapper(object):
    def __init__(self, context):
        self.context = context
        self.numbers = {}
        self.translate = {None: None}
        self.num_count = 0

    def __getitem__(self, key):
        if key is None:
            return None
        elif key < 0:
            value = self.numbers.get(key)
            if value is None:
                self.num_count -= 1
                value = self.num_count
                self.numbers[key] = value
            self.translate[value] = key
        else:
            value = self.context[key]
            if value is None or type(value) == int:
                if value is not None:
                    key = value
                    okey = value
                else:
                    okey = None
                value = self.numbers.get(key)
                if value is None:
                    self.num_count -= 1
                    value = self.num_count
                    self.numbers[key] = value
                self.translate[value] = okey

        if not is_variable(value):
            value = value.apply(self)
        return value


def substitute_call_args(terms, context):
    """

    :param terms:
    :param context:
    :return:
    """
    result = []
    cw = _ContextWrapper(context)
    for term in terms:
        if term is None:
            cw.num_count -= 1
            result.append(cw.num_count)
        elif type(term) == int:
            v = cw[term]
            result.append(v)
        else:
            result.append(term.apply(cw))
    return result, cw.translate


def substitute_head_args(terms, context):
    """
    Extract the clause head arguments from the clause context.
    :param terms: head arguments. These can contain variables >0.
    :param context: clause context. These can contain variable <0.
    :return: input terms where variables are substituted by their values in the context
    """
    result = []
    for term in terms:
        result.append(substitute_simple(term, context))
    return result


def substitute_simple(term, context):
    """

    :param term:
    :param context:
    :return:
    """
    if term is None:
        return None
    elif type(term) == int:
        return context[term]
    else:
        return term.apply(context)


def unify_call_head(call_args, head_args, target_context):
    """
    Unify argument list from clause call and clause head.
    :param call_args: arguments of the call
    :param head_args: arguments of the head
    :param target_context: list of values of variables in the clause
    :raise UnifyError: unification failed
    """
    source_values = {}  # contains the values unified to the variables in the call arguments
    for call_arg, head_arg in zip(call_args, head_args):
        _unify_call_head_single(call_arg, head_arg, target_context, source_values)
    result = substitute_all(target_context, source_values)
    return result


def _unify_call_head_single(source_value, target_value, target_context, source_values):
    """
    Unify a call argument with a clause head argument.
    :param source_value: value occuring in clause call
    :type source_value: Term or variable. All variables are represented as negative numbers.
    :param target_value: value occuring in clause head
    :type target_value: Term or variable. All variables are represented as positive numbers \
    corresponding to positions
    in target_context.
    :param target_context: values of the variables in the current context. Output argument.
    :param source_values:
    :raise UnifyError: unification failed

    The values in the target_context contain only variables from the source context \
    (i.e. negative numbers) (or Terms).
    Initially values are set to None.
    """
    if is_variable(target_value):  # target_value is variable (integer >= 0)
        assert type(target_value) == int and target_value >= 0
        target_context[target_value] = \
            unify_value(source_value, target_context[target_value], source_values)
    else:  # target_value is a Term (which can still contain variables)
        if is_variable(source_value):  # source value is variable (integer < 0)
            if source_value is None:
                pass
            else:
                assert type(source_value) == int and source_value < 0
                # This means that *all* occurrences of source_value should unify with the same value.
                # We keep track of the value for each source_value, and unify the target_value with the current value.
                # Note that we unify two values in the same scope.
                source_values[source_value] = \
                    unify_value(source_values.get(source_value), target_value, source_values)
        else:  # source value is a Term (which can still contain variables)
            if target_value.signature == source_value.signature:
                # When signatures match, recursively unify the arguments.
                for s_arg, t_arg in zip(source_value.args, target_value.args):
                    _unify_call_head_single(s_arg, t_arg, target_context, source_values)
            else:
                raise UnifyError()


class _VarTranslateWrapper(object):
    def __init__(self, var_translate, min_var):
        self.base = var_translate
        self.min_var = min_var

    def __getitem__(self, item):
        if item in self.base:
            return self.base[item]
        else:
            self.min_var -= 1
            self.base[item] = self.min_var
            return self.min_var


def unify_call_return(call_args, head_args, target_context, var_translate, min_var, mask=None):
    """
    Unify argument list from clause call and clause head.
    :param call_args: arguments of the call
    :param head_args: arguments of the head
    :param target_context: list of values of variables in the clause
    :param var_translate:
    :raise UnifyError: unification failed
    """
    if mask is None:
        mask = [True] * len(call_args)

    var_translate = _VarTranslateWrapper(var_translate, min_var)
    source_values = {}  # contains the values unified to the variables in the call arguments
    for call_arg, head_arg, mask_arg in zip(call_args, head_args, mask):
        if mask_arg:
            # Translate the variables in the source value (V3) to negative variables in current context (V2)
            call_arg = substitute_simple(call_arg, var_translate)
            _unify_call_return_single(call_arg, head_arg, target_context, source_values)
    result = substitute_all(target_context, source_values)
    return result


def _unify_call_return_single(source_value, target_value, target_context, source_values):
    """

    :param source_value: value from call result. Can contain variables < 0.
    :param target_value: value from call itself. Can contain variables >= 0.
    :param target_context:
    :param source_values:
    :return:
    """
    if is_variable(target_value):
        if target_value is None:
            pass
        elif target_value >= 0:
            # Get current value of target.
            current_value = target_context[target_value]
            # Three possibilities: None, negative variable (from V2), Term (with variables from V2)
            if current_value is None:
                target_context[target_value] = source_value
            elif type(current_value) == int:
                # Negative variable (from V2)
                source_values[current_value] = unify_value(
                    source_values.get(current_value), source_value, source_values)
            else:
                # Term (with variables from V2)
                target_context[target_value] = unify_value(
                    current_value, source_value, source_values)
        else:
            # Not possible. Target comes from static variables
            assert not target_value < 0
    else:
        # Target value is Term (which can contain variables)
        if is_variable(source_value):  # source value is variable (integer < 0)
            assert type(source_value) == int and source_value < 0
            source_values[source_value] = \
                unify_value(source_values.get(source_value), target_value, source_values)
        else:  # source value is a Term (which can still contain variables)
            if target_value.signature == source_value.signature:
                # When signatures match, recursively unify the arguments.
                for s_arg, t_arg in zip(source_value.args, target_value.args):
                    _unify_call_return_single(s_arg, t_arg, target_context, source_values)
            else:
                raise UnifyError()



#! /usr/bin/env python

from __future__ import print_function
import sys

version = '2.1.0.9'

if len(sys.argv) == 1:
    from problog import setup as problog_setup
    problog_setup.install()
else:    

    from setuptools import setup, find_packages
    from setuptools.command.install import install


    class ProbLogInstall(install):
        def run(self):
            install.run(self)
            from problog import setup as problog_setup
            try:
                problog_setup.install()
            except Exception as err:
                print ('Optional ProbLog installation failed: %s' % err, file=sys.stderr)

    package_data = {
        'problog': [
            'bin/darwin/dsharp', 
            'bin/darwin/maxsatz', 
            'bin/linux/dsharp', 
            'bin/linux/maxsatz',
            'bin/windows/dsharp.exe',
            'bin/windows/maxsatz.exe',
            'bin/windows/libgcc_s_dw2-1.dll',
            'bin/windows/libstdc++-6.dll',
            'lib/sdd/*.h',
            'lib/sdd/*.c',
            'lib/sdd/linux/libsdd.so',
            'lib/sdd/darwin/libsdd.a',
            'lib/sdd/*.py',
            'web/*.py',
            'web/editor_local.html',
            'web/editor_adv.html',
            'web/js/problog_editor.js',
            'library/lists.pl'
        ]
    }


    setup(
        name='problog',
        version=version,
        description='ProbLog2: Probabilistic Logic Programming toolbox',
        url='https://dtai.cs.kuleuven.be/problog',
        author='ProbLog team',
        author_email='anton.dries@cs.kuleuven.be',
        license='Apache Software License',
        classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: Apple Public Source License',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
        ],
        keywords='prolog probabilistic logic',
        packages=find_packages(),
        entry_points={
            'console_scripts': ['problog=problog.tasks:main']       
        },
        package_data=package_data,
        cmdclass={
            'install': ProbLogInstall
        }
    )


